/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER
/// and THE UNIVERSITY OF MANCHESTER, seanb@cs.man.ac.uk

////////////////////////////////////////////////////////////////////////////
/// FaCT Client                                                          ///
/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER                     ///
/// and THE UNIVERSITY OF MANCHESTER                                     /// 
///					                                 ///
/// This program is free software; you can redistribute it and/or        ///
/// modify it under the terms of the GNU General Public License          ///
/// as published by the Free Software Foundation; either version 2       ///
/// of the License, or (at your option) any later version.               ///
///                                                                      ///
/// This program is distributed in the hope that it will be useful,      ///
/// but WITHOUT ANY WARRANTY; without even the implied warranty of       ///
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        ///
/// GNU General Public License for more details.                         ///
///                                                                      ///
/// You should have received a copy of the GNU General Public License    ///
/// along with this program; if not, write to the Free Software          ///
/// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            ///
/// 02111-1307, USA.                                                     ///
///                                                                      ///
/// Enquiries about FaCT should be directed to:                          ///
///                                                                      ///
/// email: horrocks@cs.man.ac.uk                                         ///
/// www:   http://www.cs.man.ac.uk/~horrocks                             ///
/// smail: Ian Horrocks, Department of Computer Science, Oxford Road,    ///
///       Manchester 467M13 9PL, United Kingdom                          ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
///                                                                      ///
/// This class is intended to demonstrate how to write a Java client     ///
/// that will communicate with the FaCT description logic reasoner via   ///
/// the CORBA-based interface. It is not supported or maintained and no  ///
/// guarantees are supplied as to the quality or robustness of the code. ///
/// SKB, 4/10/99                                                         ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

package img.fact.client;

import java.util.*;
import java.io.*;
import java.net.*;
import img.fact.*;

import org.w3c.dom.*;
import org.xml.sax.*;
import org.apache.xerces.dom.*;
import org.apache.xml.serialize.*;

/** A simple little command line compiler that takes a knowledge base in XML 
** format and sends it off to the classifier.  
*/
public class Compiler {

private CompilerLogger logger=null;
private ClientHandler ch;
private boolean clearBeforeCompiling=true;
private static int TRANSACTION = 10001;
private static int XML = 10002;
private static int SERVER = 10003;
private static int FORMAT = 10004;
private static int FILE = 10005;
private static int UNKNOWN = 10099;

public class CompilerException extends Exception {
public int code;
public String information;
  //	constructors
public CompilerException(int c, String i) {
	super();
	code = c;
	information = i;
}
}

private void compilerError(int c, String msg) throws CompilerException {
  throw new CompilerException(c, msg);
}

public Compiler(String name) { 
  String ior;
  String ref=null;
  log("Connecting...");
  ClientConnector cc = new ClientConnector("localhost",
					   "8000",
					   "FaCTServer",
					   "compiler");
  ch = cc.connect();
}

public Compiler(String host, String port, String serverName, String name) { 
  String ior;
  String ref=null;
  log("Connecting...");
  ClientConnector cc = new ClientConnector(host,
					   port,
					   serverName,
					   name);
  ch = cc.connect();
}

public Compiler(ClientHandler c) {
  ch = c;
}

public void clearBeforeCompiling(boolean b) {
  clearBeforeCompiling = b;
}

private void release() {
  if (ch!=null) {
    ch.release();
  }
}

private Document parse(String uri) throws CompilerException {
    org.w3c.dom.Document doc = null;
    try {
	javax.xml.parsers.DocumentBuilderFactory dbf =
	    javax.xml.parsers.DocumentBuilderFactory.newInstance();
	dbf.setValidating(false);
	javax.xml.parsers.DocumentBuilder builder = dbf.newDocumentBuilder();
	
	doc = builder.parse(uri);
    } catch (Exception e) {
	compilerError(Compiler.XML, e.getMessage());
    }	
    return doc;
}

private Document parse(File f) throws CompilerException {
  try {
    String uri = f.toURL().toString();
    return parse(uri);
  } catch (MalformedURLException ex) {
    compilerError(Compiler.FILE, ex.getMessage());
  }
  return null;
}
  
public void compileInTransaction(File f) throws CompilerException {
  if (ch.begin_transaction()) {
    compile(f);
    if (ch.end_transaction()) {
      log("Compilation completed.");
    } else {
      compilerError(Compiler.TRANSACTION, "Couldn't end");
    }
  } else {
    compilerError(Compiler.TRANSACTION, "Couldn't start");
  }
}

public void compile(File f) throws CompilerException{
  Document doc = parse(f);
  compile(doc);
}

public void compile(Document doc) throws CompilerException{
  try {
    if (clearBeforeCompiling) {
      ch.clear();
    }
    Element e = doc.getDocumentElement();
    if (!e.getTagName().equals("KNOWLEDGEBASE")) {
      compilerError(Compiler.FORMAT, "No knowledgebase!");
    }
    NodeList tells = e.getElementsByTagName("DEFCONCEPT");
    log(Integer.toString(tells.getLength()) + " concepts");
    for (int i=0; i<tells.getLength(); i++) {
      Element tell = (Element) tells.item(i);
      String nm = tell.getAttribute("NAME");
      ch.defconcept(nm);
    }
    tells = e.getElementsByTagName("DEFROLE");
    log(Integer.toString(tells.getLength()) + " roles");
    for (int i=0; i<tells.getLength(); i++) {
      Element tell = (Element) tells.item(i);
      String nm = tell.getAttribute("NAME");
      ch.defrole(nm);
    }
    tells = e.getElementsByTagName("IMPLIESC");
    log(Integer.toString(tells.getLength()) + " impliesc");
    for (int i=0; i<tells.getLength(); i++) {
      /*  This is all a bit of a hack due to the fact that the file may
      contain white space in it. */ 
      
      Element tell = (Element) tells.item(i);
      tell.normalize();
      Node[] args = stripNonElement(tell.getChildNodes());
      // Assumes there are two!
      if (args.length != 2) {
 	compilerError(Compiler.FORMAT, "Wrong number of args for IMPLIESC");
      } else {
       Element ante = (Element) args[0];
       Element cons = (Element) args[1];
       ch.impliesC(xmlToString(ante),xmlToString(cons));
      }
    }
    tells = e.getElementsByTagName("EQUALC");
    log(Integer.toString(tells.getLength()) + " equalsc");
    for (int i=0; i<tells.getLength(); i++) {
      /*  This is all a bit of a hack due to the fact that the file may
      contain white space in it. */ 
      
      Element tell = (Element) tells.item(i);
      tell.normalize();
      Node[] args = stripNonElement(tell.getChildNodes());
      // Assumes there are two!
      if (args.length != 2) {
 	compilerError(Compiler.FORMAT, "Wrong number of args for EQUALC");
      } else {
       Element ante = (Element) args[0];
       Element cons = (Element) args[1];
       ch.equalC(xmlToString(ante),xmlToString(cons));
      }
    }
    tells = e.getElementsByTagName("IMPLIESR");
    log(Integer.toString(tells.getLength()) + " impliesr");
    for (int i=0; i<tells.getLength(); i++) {
      /*  This is all a bit of a hack due to the fact that the file may
      contain white space in it. */ 
      
      Element tell = (Element) tells.item(i);
      tell.normalize();
      Node[] args = stripNonElement(tell.getChildNodes());
      // Assumes there are two!
      if (args.length != 2) {
 	compilerError(Compiler.FORMAT, "Wrong number of args for IMPLIESR");
      } else {
       Element ante = (Element) args[0];
       Element cons = (Element) args[1];
       ch.impliesR(xmlToString(ante),xmlToString(cons));
      }
    }
    tells = e.getElementsByTagName("EQUALR");
    log(Integer.toString(tells.getLength()) + " equalr");
    for (int i=0; i<tells.getLength(); i++) {
      /*  This is all a bit of a hack due to the fact that the file may
      contain white space in it. */ 
      
      Element tell = (Element) tells.item(i);
      tell.normalize();
      Node[] args = stripNonElement(tell.getChildNodes());
      // Assumes there are two!
      if (args.length != 2) {
 	compilerError(Compiler.FORMAT, "Wrong number of args for EQUALR");
      } else {
       Element ante = (Element) args[0];
       Element cons = (Element) args[1];
       ch.equalR(xmlToString(ante),xmlToString(cons));
      }
    }
    tells = e.getElementsByTagName("TRANSITIVE");
    log(Integer.toString(tells.getLength()) + " transitive");
    for (int i=0; i<tells.getLength(); i++) {
      /*  This is all a bit of a hack due to the fact that the file may
      contain white space in it. */ 
      
      Element tell = (Element) tells.item(i);
      tell.normalize();
      Node[] args = stripNonElement(tell.getChildNodes());
      // Assumes there are two!
      if (args.length != 1) {
 	compilerError(Compiler.FORMAT, "Wrong number of args for TRANSITIVE");
      } else {
	Element arg = (Element) args[0];
	ch.transitive(xmlToString(arg));
      }
    }
    tells = e.getElementsByTagName("FUNCTIONAL");
    log(Integer.toString(tells.getLength()) + " functional");
    for (int i=0; i<tells.getLength(); i++) {
      /*  This is all a bit of a hack due to the fact that the file may
      contain white space in it. */ 
      
      Element tell = (Element) tells.item(i);
      tell.normalize();
      Node[] args = stripNonElement(tell.getChildNodes());
      // Assumes there are two!
      if (args.length != 1) {
 	compilerError(Compiler.FORMAT, "Wrong number of args for FUNCTIONAL");
      } else {
	Element arg = (Element) args[0];
	ch.functional(xmlToString(arg));
      }
    }
//       } else if (tell.getTagName().equals("TRANSITIVE")) {
// 	NodeList args = tell.getChildNodes();
// 	Element arg = (Element) args.item(0);
// 	if (arg==null) {
// 	  compilerError(Compiler.FORMAT, "Element not found");
// 	}
// 	ch.transitive(arg.toString());
// // 	String rn = arg.getAttribute("NAME");
// // 	ch.transitive(rn);
//       } else if (tell.getTagName().equals("FUNCTIONAL")) {
// 	NodeList args = tell.getChildNodes();
// 	Element arg = (Element) args.item(0);
// 	if (arg==null) {
// 	  compilerError(Compiler.FORMAT, "Element not found");
// 	}
// 	ch.functional(arg.toString());
// // 	String rn = arg.getAttribute("NAME");
// // 	ch.functional(rn);
//       } else {
// 	log("Ignored: " + tell.toString());
//       }

// //       long t = System.currentTimeMillis();
// //       System.out.print(" ");
// //       System.out.print(t-s);
      
//     }
  } catch (InternalException ex) {
    compilerError(Compiler.SERVER, Short.toString(ex.code) + " " + ex.information);
  } catch (TransactionRequiredException ex) {
    compilerError(Compiler.SERVER, Short.toString(ex.code) + " " + ex.information);
  } catch (OpUnimplementedException ex) {
    compilerError(Compiler.SERVER, Short.toString(ex.code) + " " + ex.information);
  } catch (ExprErrorException ex) {
    compilerError(Compiler.SERVER, Short.toString(ex.code) + " " + ex.information);
  } catch (ClassCastException ex) {
    compilerError(Compiler.XML, ex.getMessage());
    ex.printStackTrace(System.err);
  }
}

private String xmlToString(Node n) {
  StringWriter sw = new StringWriter();
  OutputFormat of = new OutputFormat();
  XMLSerializer serializer = new XMLSerializer(sw, of);
  try {
      serializer.serialize((Element) n);
  } catch (Exception e) {
      return "";
  }
//   DOMWriter dw  = new DOMWriter(new PrintWriter(sw));
//   dw.print(n);
  return sw.toString();
}

  /** Strips out any text nodes from the given node list. */
private Node[] stripNonElement(NodeList list) {
  Collection stripped = new ArrayList();
  for (int i=0; i<list.getLength(); i++) {
    try {
      Element tell = (Element) list.item(i);
      stripped.add(tell);
    } catch (ClassCastException ex) {
      // Ignore it 
    }
  }
  return (Node[]) stripped.toArray(new Node[stripped.size()]);
}

public void setLogger(CompilerLogger l) {
  logger = l;
}

private void log(String str) {
  if (logger!=null) {
    logger.compilerMessage(str);
  }
};

// public static void main(String[] args) {

//   Hashtable argSpec = new Hashtable();
//   argSpec.put("nsHost", new Integer(1));
//   argSpec.put("nsPort", new Integer(1));
//   argSpec.put("serverName", new Integer(1));
//   argSpec.put("noClear", new Integer(0));
//   argSpec.put("help", new Integer(0));
  
//   uk.ac.man.cs.img.util.CommandLineArguments cla = 
//     new uk.ac.man.cs.img.util.CommandLineArguments(argSpec);

//   // Extract the arguments
//   String serverName = "FaCTServer";
//   String nsHost = "localhost";
//   String nsPort = "8000";
//   boolean clearBefore = true;
//   String fileName = "";
//   try {
//     String[] tmp;
//     cla.setArgs(args);
//     if (cla.flagPresent("help")) {
//       printHelp();
//       System.exit(0);
//     }
//     tmp = cla.argsForSelector("serverName");
//     if (tmp!=null) {
//       serverName = tmp[0];
//     }
//     tmp = cla.argsForSelector("nsHost");
//     if (tmp!=null) {
//       nsHost = tmp[0];
//     }
//     tmp = cla.argsForSelector("nsPort");
//     if (tmp!=null) {
//       nsPort = tmp[0];
//     }
//     clearBefore = !cla.flagPresent("noClear");
//     tmp = cla.otherArgs();
//     fileName = tmp[0];
//   } catch (Exception ex) {
//     System.err.println("Usage: java uk.ac.man.cs.img.fact.client.Compiler [-help] [-nsHost host] [-nsPort port] [-serverName name] [-noClear] file");
//     System.exit(1);
//   }

//   Compiler c = new Compiler(nsHost, nsPort, serverName, "compiler");
//   c.clearBeforeCompiling(clearBefore);
//   c.setLogger(new CompilerLogger() {
//     public void compilerMessage(String msg) {
//       System.out.println(msg);
//     }
//   });
//   try {
//     c.compileInTransaction(new File(fileName));
//     c.release();
//   } catch (CompilerException ex) {
//     c.release();
//     System.err.println("ERROR: " + ex.code + " " +
// 		       ex.information);
//     ex.printStackTrace(System.err);
//     System.exit(1);
//   } catch (Exception ex) {
//     c.release();
//     System.err.println("ERROR: " + ex.toString());
//     ex.printStackTrace(System.err);
//     System.exit(1);
//   }
// }

public static void printHelp() {
  System.out.println("Usage: java uk.ac.man.cs.img.fact.client.Compiler [options] file");
  System.out.println("Compiles an xml knowledge base");
  System.out.println("\t-nsHost\t\tHost running naming service");
  System.out.println("\t-nsPort\t\tPort running naming service");
  System.out.println("\t-serverName\t\tName the server is bound to");
  System.out.println("\t-noClear\t\tDon't clear the knowledge base before compiling");
  System.out.println("\t-help\t\tPrint this message");
}

}
