package img.fact.client;

import java.util.*;
import java.io.*;
import java.net.*;
import lispwrapper.*;
import img.fact.sexpr.*;
import img.fact.xml.*;
import img.fact.*;

import org.w3c.dom.*;
import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;

import com.sun.xml.tree.*;
import com.sun.xml.parser.Resolver;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import att.grappa.*;

/** A simple little class that produces a graph of a FaCT
 * model that can be read by graphing applications
 */
public class GrappaGrapher {

private Classifier cl;
private ClientHandler ch;

private XMLToName xtn;

public GrappaGrapher() { 
  String ior;
  String ref=null;
  ClientConnector cc = new ClientConnector("localhost",
					   "8000",
					   "FaCTServer",
					   "grapher");
  ch = cc.connect();
  initializeXML();
}

public GrappaGrapher(String host, String port, String serverName, String name) { 
  String ior;
  String ref=null;
  ClientConnector cc = new ClientConnector(host,
					   port,
					   serverName,
					   name);
  ch = cc.connect();
  initializeXML();
}

public GrappaGrapher(ClientHandler c) {
  ch = c;
  initializeXML();
}

public void initializeXML() {
  try {
    xtn = new XMLToName();
  } catch (XMLConversionException e) {
    System.err.println("Error: " + e.getMessage());
    e.printStackTrace();
    System.exit(99);
  }
}

public void release() {
  if (ch!=null) {
    ch.release();
  }
}

public void graph(String top, PrintStream os) throws XMLConversionException {
  printHeader(os);
  if (top.equals(":TOP")) {
    top = "<TOP/>";
  } else {
    top = "<PRIMITIVE NAME=\"" + top + "\"/>";
  }
  graphConcept(top, os, true);
  graphConceptUp(top,os);
  try {
    String [][] res = ch.allSubsC(top);
    for (int i=0; i<res.length; i++) {
      graphConcept(res[i][0],os, false);
    }
    res = ch.allSupersC(top);
    for (int i=0; i<res.length; i++) {
      graphConceptUp(res[i][0],os);
    }
    printFooter(os);
  } catch (InternalException ex) {
    report (Short.toString(ex.code) + " " + ex.information);
  } catch (KBModifiedException ex) {
    report (Short.toString(ex.code) + " " + ex.information);
  } catch (OpUnimplementedException ex) {
    report (Short.toString(ex.code) + " " + ex.information);
  } catch (ExprErrorException ex) {
    report (Short.toString(ex.code) + " " + ex.information);
  }
}

private void printHeader(PrintStream os) {
  os.println("digraph generated {");
  //  os.println("  size=\"7,10\";");
  os.println("  ratio=fill;");
  os.println("  node [style=filled];");
  os.println("  node [color=pink];");
  os.println("  node [fontcolor=black];");
  os.println("  node [fontsize=12];");
}

private void printFooter(PrintStream os) {
  os.println("}");
}

private void graphConcept(String con, PrintStream os, boolean isFocus) throws XMLConversionException {
  if (!con.equals("<BOTTOM/>")) {
    String name = "";
    try {
      name = xtn.convert(con);
    } catch (XMLConversionException e) {
      System.err.println(con);
      throw e;
    }
    os.println();
    if (name.equals(":TOP")) {
      name = "TOP";
    };
    if (isFocus) {
      os.println("  " + name.toLowerCase() + "[color=pink];");
    } else {
      os.println("  " + name.toLowerCase() + "[color=green];");
    }      
    os.print("  " + name.toLowerCase() + "-> {");
    try {
      // Find all the subs, and add a link
      String [][] res = ch.directSubsC(con);
      for (int i=0; i<res.length; i++) {
	String bt = xtn.convert(res[i][0]);
	if (!bt.equals(":BOTTOM")) {
	  os.print(bt.toLowerCase() + "; ");
	}
      }
      os.println("};");
    } catch (InternalException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (KBModifiedException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (OpUnimplementedException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (ExprErrorException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    }
  }
}

private void graphConceptUp(String con, PrintStream os) throws XMLConversionException {
  if (!con.equals("<BOTTOM/>")) {
    String name = "";
    try {
      name = xtn.convert(con);
    } catch (XMLConversionException e) {
      System.err.println(con);
      throw e;
    }
    os.println();
    if (name.equals(":TOP")) {
      name = "TOP";
    };
    try {
      // Find all the subs, and add a link
      String [][] res = ch.directSupersC(con);
      for (int i=0; i<res.length; i++) {
	String bt = xtn.convert(res[i][0]);
	if (!bt.equals(":BOTTOM")) {
	  if (bt.equals(":TOP")) {
	    bt = "TOP";
	  }
	  os.println("  " + bt.toLowerCase() + " [color=cyan];");
	  os.print("  " + bt.toLowerCase() + " -> ");
	  os.println(name.toLowerCase() + ";");
	}
      }
    } catch (InternalException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (KBModifiedException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (OpUnimplementedException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (ExprErrorException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    }
  }
}

private void report(String msg) {
  System.err.println(msg);
}

public static void main(String[] args) {

  Hashtable argSpec = new Hashtable();
  argSpec.put("nsHost", new Integer(1));
  argSpec.put("nsPort", new Integer(1));
  argSpec.put("serverName", new Integer(1));
  argSpec.put("top", new Integer(1));
  argSpec.put("help", new Integer(0));
  
  img.util.CommandLineArguments cla = 
    new img.util.CommandLineArguments(argSpec);

  // Extract the arguments
  String serverName = "FaCTServer";
  String nsHost = "localhost";
  String nsPort = "8000";
  String fileName = "";
  String top = ":TOP";
  try {
    String[] tmp;
    cla.setArgs(args);
    if (cla.flagPresent("help")) {
      printHelp();
      System.exit(0);
    }
    tmp = cla.argsForSelector("serverName");
    if (tmp!=null) {
      serverName = tmp[0];
    }
    tmp = cla.argsForSelector("nsHost");
    if (tmp!=null) {
      nsHost = tmp[0];
    }
    tmp = cla.argsForSelector("nsPort");
    if (tmp!=null) {
      nsPort = tmp[0];
    }
    tmp = cla.argsForSelector("top");
    if (tmp!=null) {
      top = tmp[0];
    } else {
      top = ":TOP";
    }
    tmp = cla.otherArgs();
    //    fileName = tmp[0];
    int i = (int) (Math.random() * 1000000);
    boolean exists = true;
    while (exists) {
      i = (int) (Math.random() * 1000000);
      fileName = "_" + Integer.toString(i) + ".tmp";
      File f = new File(fileName);
      exists = f.exists();
    }
  } catch (Exception ex) {
    System.err.println("Usage: java img.fact.client.GrappaGrapher [-help] [-nsHost host] [-nsPort port] [-serverName name] [-top name]");
    System.exit(1);
  }

  GrappaGrapher dg = new GrappaGrapher(nsHost, nsPort, serverName, "grapher");
  try {

    PrintStream os = new PrintStream(new FileOutputStream(fileName));
    dg.graph(top, os);
    dg.release();
    String command = System.getProperty("dot",
					"D:\\progra~1\\graphviz\\bin\\dot.exe"); 
    command = command + " " + fileName;
    Process proc = Runtime.getRuntime().exec(command);
    InputStream pIn = proc.getInputStream();
    dg.doDemo(pIn);
    File f = new File(fileName);
    f.deleteOnExit();
  } catch (Exception ex) {
    ex.printStackTrace();
    dg.release();
    System.err.println("ERROR: " + ex.toString());
    System.exit(1);
  }

//   try {
//     Process proc = Runtime.getRuntime().exec("dot.bat");
//     OutputStream pOut = proc.getOutputStream();
//     InputStream pIn = proc.getInputStream();
//     dg.doDemo(pIn);
//   } catch (Exception ex) {
//     ex.printStackTrace();
//     System.exit(1);
//   }
}

public static void printHelp() {
  System.out.println("Usage: java img.fact.client [options] file");
  System.out.println("Produces a static dump of the hierarchy as a graph");
  System.out.println("\t-nsHost\t\tHost running naming service");
  System.out.println("\t-nsPort\t\tPort running naming service");
  System.out.println("\t-serverName\t\tName the server is bound to");
  System.out.println("\t-top\t\tTop concept to start from");
  System.out.println("\t-help\t\tPrint this message");
}

  public DemoFrame  frame  = null;
  public BirdFrame  birdframe  = null;

  public Subgraph subg = null;
  
  /*
   * If no arguments, reads from stdin; otherwise argument is input file.
   */

  void doDemo(InputStream input) {
    att.grappa.Parser program = new att.grappa.Parser(input,System.err);
    try {
      //program.debug_parse(4);
      program.parse();
    } catch(Exception ex) {
      System.err.println("Exception: " + ex.getMessage());
      ex.printStackTrace(System.err);
      System.exit(1);
    }
    Graph graph = program.getGraph();
	// for huale's version
	// graph.addToBBox(null);

    System.err.println("The graph contains " + graph.countOfElements(Grappa.NODE|Grappa.EDGE|Grappa.SUBGRAPH) + " elements.");

    graph.setEditable(true);
    graph.setMenuable(true);
    graph.setErrorWriter(new PrintWriter(System.err,true));
    //graph.printGraph(new PrintWriter(System.out));

      try {
	graph.buildObjects();
      } catch(InstantiationException ie) {
	System.err.println("Cannot instantiate: " + ie.getMessage());
	ie.printStackTrace(System.err);
	System.exit(1);
      }
    //System.err.println("bbox=" + graph.getBounds().toString());

    frame = new DemoFrame(graph);
    frame.addComponentListener(frame);

    if(subg != null) frame.getDrawPane().setupPane(subg);

    frame.show();

    birdframe = new BirdFrame(graph);
  }

  public void showBirdFrame(boolean makeVisible) {
    birdframe.setVisible(makeVisible);
  }

  private static final int initSize = 300;
  class BirdFrame implements ComponentListener
  {
    DrawPane pane = null;
    Graph graph = null;
    Frame bframe = null;
  
    public BirdFrame(Graph graph) {
      this.graph = graph;
    }

    private void buildFrame() {
      bframe = new Frame("BirdFrame");
      Dimension sz = null;
      Rectangle bnds = graph.getBounds();
      int mxsd = 10 + ((bnds.width > bnds.height) ? bnds.width : bnds.height);
      if(mxsd < initSize) {
	sz = new Dimension(bnds.width,bnds.height);
      } else {
	double factor = (double)initSize / (double)mxsd;
	sz = new Dimension((int)Math.round(factor*(double)bnds.width),(int)Math.round(factor*(double)bnds.height));
      }
      bframe.add("Center", pane = new DrawPane(graph,true,ScrollPane.SCROLLBARS_NEVER,sz));
      bframe.pack();
	// huale's version cannot handle the next line
      bframe.setSize(sz);
      bframe.addComponentListener(this);
    }

    public void setVisible(boolean viz) {
      if(bframe == null) buildFrame();
      bframe.setVisible(viz);
    }

    public DrawPane getDrawPane() {
      if(bframe == null) buildFrame();
      return pane;
    }

    public void componentResized(ComponentEvent evt) {
      //System.err.println("bird resized");
      //getDrawPane().setupPane();
      //if(bframe.isVisible()) getDrawPane().paintCanvas();
    }

    public void componentShown(ComponentEvent evt) {
      //System.err.println("bird shown");
      //getDrawPane().paintCanvas();
    }

    public void componentHidden(ComponentEvent evt) {
      //System.err.println("bird hidden");
      return;
    }

    public void componentMoved(ComponentEvent evt) {
      //System.err.println("bird moved");
      return;
    }
  }

  class DemoFrame extends Frame implements ActionListener, ComponentListener, ItemListener
  {
    Graphics gr = null;
    DrawPane pane = null;
    //    Button load = null;
    Checkbox birdseye = null;
    Button quit = null;
    Button reload = null;
    Graph graph = null;
    Panel panel = null;
  
    public DemoFrame(Graph graph) {
      super("DemoFrame");
      this.graph = graph;

      GridBagLayout gbl = new GridBagLayout();
      GridBagConstraints gbc = new GridBagConstraints();

      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gbc.fill = GridBagConstraints.HORIZONTAL;
      gbc.anchor = GridBagConstraints.NORTHWEST;

      panel = new Panel();
      panel.setLayout(gbl);

      birdseye = new Checkbox("Bird's Eye");
      gbl.setConstraints(birdseye,gbc);
      panel.add(birdseye);
      birdseye.addItemListener(this);

      quit = new Button("Quit");
      gbl.setConstraints(quit,gbc);
      panel.add(quit);
      quit.addActionListener(this);

      //      add("Center", pane = new DrawPane(graph,false,1,new Dimension(650,650)));
      add("Center", pane = new DrawPane(graph,false,ScrollPane.SCROLLBARS_AS_NEEDED,new Dimension(650,650)));
      add("West", panel);

      pack();
      setSize(600,400);
      doLayout();
      pane.getCanvas().setMinimumSize(pane.getViewportSize());
    }

    public DrawPane getDrawPane() {
      return pane;
    }

    public void itemStateChanged(ItemEvent evt) {
      if(evt.getSource() instanceof Checkbox) {
	Checkbox tgt = (Checkbox)evt.getSource();
	if(tgt == birdseye) {
	  showBirdFrame(birdseye.getState());
	}
      }
    }

    public void actionPerformed(ActionEvent evt) {
      if(evt.getSource() instanceof Button) {
	Button tgt = (Button)evt.getSource();
	if(tgt == quit) {
	  System.exit(0);
	}
      }
    }

    public void componentResized(ComponentEvent evt) {
      //System.err.println("frame resized");
      //pane.setupPane();
      //pane.paintCanvas();
    }

    public void componentShown(ComponentEvent evt) {
      //System.err.println("frame shown");
      //pane.paintCanvas();
    }

    public void componentHidden(ComponentEvent evt) {
      //System.err.println("frame hidden");
      return;
    }

    public void componentMoved(ComponentEvent evt) {
      //System.err.println("frame moved");
      return;
    }
  }



}
